@@ -1,96 +1,103 @@
 use std::collections::HashMap;
+use std::sync::atomic::Ordering;
 use std::sync::Arc;
 
+use indexmap::IndexMap;
+use metrics::{Counter, Gauge, Histogram, Key, KeyName, Recorder, Unit};
+use metrics_util::recency::{GenerationalPrimitives, Recency};
+use metrics_util::Registry;
 use parking_lot::RwLock;
 
-use metrics::{GaugeValue, Key, Recorder, Unit};
-use metrics_util::{Handle, MetricKind, Recency, Registry};
-
-use crate::common::{sanitize_key_name, Snapshot};
+use crate::common::{
+    sanitize_description, sanitize_label_key, sanitize_label_value, sanitize_metric_name, Snapshot,
+};
 use crate::distribution::{Distribution, DistributionBuilder};
 
 pub(crate) struct Inner {
-    pub registry: Registry<Key, Handle>,
-    pub recency: Recency<Key>,
-    pub distributions: RwLock<HashMap<String, HashMap<Vec<String>, Distribution>>>,
+    pub registry: Registry<GenerationalPrimitives>,
+    pub recency: Recency,
+    pub distributions: RwLock<HashMap<String, IndexMap<Vec<String>, Distribution>>>,
     pub distribution_builder: DistributionBuilder,
     pub descriptions: RwLock<HashMap<String, &'static str>>,
-    pub global_labels: HashMap<String, String>,
+    pub global_labels: IndexMap<String, String>,
 }
 
 impl Inner {
-    pub fn registry(&self) -> &Registry<Key, Handle> {
-        &self.registry
-    }
-
     fn get_recent_metrics(&self) -> Snapshot {
-        let metrics = self.registry.get_handles();
-
         let mut counters = HashMap::new();
-        let mut gauges = HashMap::new();
-
-        for ((kind, key), (gen, handle)) in metrics.into_iter() {
-            if !self.recency.should_store(kind, &key, gen, self.registry()) {
+        let counter_handles = self.registry.get_counter_handles();
+        for (key, counter) in counter_handles {
+            let gen = counter.get_generation();
+            if !self.recency.should_store_counter(&key, gen, &self.registry) {
                 continue;
             }
 
-            match kind {
-                MetricKind::Counter => {
-                    let value = handle.read_counter();
+            let (name, labels) = key_to_parts(&key, &self.global_labels);
+            let value = counter.get_inner().load(Ordering::Acquire);
+            let entry =
+                counters.entry(name).or_insert_with(HashMap::new).entry(labels).or_insert(0);
+            *entry = value;
+        }
 
-                    let (name, labels) = key_to_parts(&key, &self.global_labels);
-                    let entry = counters
-                        .entry(name)
-                        .or_insert_with(HashMap::new)
-                        .entry(labels)
-                        .or_insert(0);
-                    *entry = value;
-                }
-                MetricKind::Gauge => {
-                    let value = handle.read_gauge();
-
-                    let (name, labels) = key_to_parts(&key, &self.global_labels);
-                    let entry = gauges
-                        .entry(name)
-                        .or_insert_with(HashMap::new)
-                        .entry(labels)
-                        .or_insert(0.0);
-                    *entry = value;
-                }
-                MetricKind::Histogram => {
-                    let (name, labels) = key_to_parts(&key, &self.global_labels);
-
-                    let mut wg = self.distributions.write();
-                    let entry = wg
-                        .entry(name.clone())
-                        .or_insert_with(HashMap::new)
-                        .entry(labels)
-                        .or_insert_with(|| {
-                            self.distribution_builder
-                                .get_distribution(name.as_str())
-                                .expect("failed to create distribution")
-                        });
-
-                    handle.read_histogram_with_clear(|samples| entry.record_samples(samples));
-                }
+        let mut gauges = HashMap::new();
+        let gauge_handles = self.registry.get_gauge_handles();
+        for (key, gauge) in gauge_handles {
+            let gen = gauge.get_generation();
+            if !self.recency.should_store_gauge(&key, gen, &self.registry) {
+                continue;
             }
+
+            let (name, labels) = key_to_parts(&key, &self.global_labels);
+            let value = f64::from_bits(gauge.get_inner().load(Ordering::Acquire));
+            let entry =
+                gauges.entry(name).or_insert_with(HashMap::new).entry(labels).or_insert(0.0);
+            *entry = value;
+        }
+
+        let histogram_handles = self.registry.get_histogram_handles();
+        for (key, histogram) in histogram_handles {
+            let gen = histogram.get_generation();
+            if !self.recency.should_store_histogram(&key, gen, &self.registry) {
+                // Since we store aggregated distributions directly, when we're told that a metric
+                // is not recent enough and should be/was deleted from the registry, we also need to
+                // delete it on our side as well.
+                let (name, labels) = key_to_parts(&key, &self.global_labels);
+                let mut wg = self.distributions.write();
+                let delete_by_name = if let Some(by_name) = wg.get_mut(&name) {
+                    by_name.remove(&labels);
+                    by_name.is_empty()
+                } else {
+                    false
+                };
+
+                // If there's no more variants in the per-metric-name distribution map, then delete
+                // it entirely, otherwise we end up with weird empty output during render.
+                if delete_by_name {
+                    wg.remove(&name);
+                }
+
+                continue;
+            }
+
+            let (name, labels) = key_to_parts(&key, &self.global_labels);
+
+            let mut wg = self.distributions.write();
+            let entry = wg
+                .entry(name.clone())
+                .or_insert_with(IndexMap::new)
+                .entry(labels)
+                .or_insert_with(|| self.distribution_builder.get_distribution(name.as_str()));
+
+            histogram.get_inner().clear_with(|samples| entry.record_samples(samples));
         }
 
         let distributions = self.distributions.read().clone();
 
-        Snapshot {
-            counters,
-            gauges,
-            distributions,
-        }
+        Snapshot { counters, gauges, distributions }
     }
 
-    pub fn render(&self) -> String {
-        let Snapshot {
-            mut counters,
-            mut distributions,
-            mut gauges,
-        } = self.get_recent_metrics();
+    fn render(&self) -> String {
+        let Snapshot { mut counters, mut distributions, mut gauges } = self.get_recent_metrics();
 
         let mut output = String::new();
         let descriptions = self.descriptions.read();
@@ -124,11 +131,9 @@ impl Inner {
                 write_help_line(&mut output, name.as_str(), desc);
             }
 
-            let distribution_type = self
-                .distribution_builder
-                .get_distribution_type(name.as_str());
+            let distribution_type = self.distribution_builder.get_distribution_type(name.as_str());
             write_type_line(&mut output, name.as_str(), distribution_type);
-            for (labels, distribution) in by_labels.drain() {
+            for (labels, distribution) in by_labels.drain(..) {
                 let (sum, count) = match distribution {
                     Distribution::Summary(summary, quantiles, sum) => {
                         for quantile in quantiles.iter() {
@@ -192,7 +197,9 @@ impl Inner {
 /// This recorder should be composed with other recorders or installed globally via
 /// [`metrics::set_boxed_recorder`].
 ///
-///
+/// Most users will not need to interact directly with the recorder, and can simply deal with the
+/// builder methods on [`PrometheusBuilder`](crate::PrometheusBuilder) for building and installing
+/// the recorder/exporter.
 pub struct PrometheusRecorder {
     inner: Arc<Inner>,
 }
@@ -200,116 +207,81 @@ pub struct PrometheusRecorder {
 impl PrometheusRecorder {
     /// Gets a [`PrometheusHandle`] to this recorder.
     pub fn handle(&self) -> PrometheusHandle {
-        PrometheusHandle {
-            inner: self.inner.clone(),
-        }
+        PrometheusHandle { inner: self.inner.clone() }
     }
 
-    fn add_description_if_missing(&self, key: &Key, description: Option<&'static str>) {
-        if let Some(description) = description {
-            let mut descriptions = self.inner.descriptions.write();
-            if !descriptions.contains_key(key.name().to_string().as_str()) {
-                descriptions.insert(key.name().to_string(), description);
-            }
-        }
+    fn add_description_if_missing(&self, key_name: KeyName, description: &'static str) {
+        let sanitized = sanitize_metric_name(key_name.as_str());
+        let mut descriptions = self.inner.descriptions.write();
+        descriptions.entry(sanitized).or_insert(description);
     }
 }
 
 impl From<Inner> for PrometheusRecorder {
     fn from(inner: Inner) -> Self {
-        PrometheusRecorder {
-            inner: Arc::new(inner),
-        }
+        PrometheusRecorder { inner: Arc::new(inner) }
     }
 }
 
 impl Recorder for PrometheusRecorder {
-    fn register_counter(&self, key: &Key, _unit: Option<Unit>, description: Option<&'static str>) {
-        self.add_description_if_missing(&key, description);
-        self.inner
-            .registry()
-            .op(MetricKind::Counter, key, |_| {}, Handle::counter);
+    fn describe_counter(&self, key_name: KeyName, _unit: Option<Unit>, description: &'static str) {
+        self.add_description_if_missing(key_name, description);
     }
 
-    fn register_gauge(&self, key: &Key, _unit: Option<Unit>, description: Option<&'static str>) {
-        self.add_description_if_missing(&key, description);
-        self.inner
-            .registry()
-            .op(MetricKind::Gauge, key, |_| {}, Handle::gauge);
+    fn describe_gauge(&self, key_name: KeyName, _unit: Option<Unit>, description: &'static str) {
+        self.add_description_if_missing(key_name, description);
     }
 
-    fn register_histogram(
+    fn describe_histogram(
         &self,
-        key: &Key,
+        key_name: KeyName,
         _unit: Option<Unit>,
-        description: Option<&'static str>,
+        description: &'static str,
     ) {
-        self.add_description_if_missing(&key, description);
-        self.inner
-            .registry()
-            .op(MetricKind::Histogram, key, |_| {}, Handle::histogram);
+        self.add_description_if_missing(key_name, description);
     }
 
-    fn increment_counter(&self, key: &Key, value: u64) {
-        self.inner.registry().op(
-            MetricKind::Counter,
-            key,
-            |h| h.increment_counter(value),
-            Handle::counter,
-        );
+    fn register_counter(&self, key: &Key) -> Counter {
+        self.inner.registry.get_or_create_counter(key, |c| c.clone().into())
     }
 
-    fn update_gauge(&self, key: &Key, value: GaugeValue) {
-        self.inner.registry().op(
-            MetricKind::Gauge,
-            key,
-            |h| h.update_gauge(value),
-            Handle::gauge,
-        );
+    fn register_gauge(&self, key: &Key) -> Gauge {
+        self.inner.registry.get_or_create_gauge(key, |c| c.clone().into())
     }
 
-    fn record_histogram(&self, key: &Key, value: f64) {
-        self.inner.registry().op(
-            MetricKind::Histogram,
-            key,
-            |h| h.record_histogram(value),
-            Handle::histogram,
-        );
+    fn register_histogram(&self, key: &Key) -> Histogram {
+        self.inner.registry.get_or_create_histogram(key, |c| c.clone().into())
     }
 }
 
-/// Handle to [`PrometheusRecorder`].
+/// Handle for accessing metrics stored via [`PrometheusRecorder`].
 ///
-/// Useful for exposing a scrape endpoint on an existing HTTP/HTTPS server.
+/// In certain scenarios, it may be necessary to directly handle requests that would otherwise be
+/// handled directly by the HTTP listener, or push gateway background task.  [`PrometheusHandle`]
+/// allows rendering a snapshot of the current metrics stored by an installed [`PrometheusRecorder`]
+/// as a payload conforming to the Prometheus exposition format.
 #[derive(Clone)]
 pub struct PrometheusHandle {
     inner: Arc<Inner>,
 }
 
 impl PrometheusHandle {
-    /// Returns the metrics in Prometheus accepted String format.
+    /// Takes a snapshot of the metrics held by the recorder and generates a payload conforming to
+    /// the Prometheus exposition format.
     pub fn render(&self) -> String {
         self.inner.render()
     }
 }
 
-fn key_to_parts(key: &Key, defaults: &HashMap<String, String>) -> (String, Vec<String>) {
-    let name = sanitize_key_name(key.name());
+fn key_to_parts(key: &Key, defaults: &IndexMap<String, String>) -> (String, Vec<String>) {
+    let name = sanitize_metric_name(key.name());
     let mut values = defaults.clone();
     key.labels().into_iter().for_each(|label| {
-        values.insert(label.key().into(), label.value().into());
+        values.insert(label.key().to_string(), label.value().to_string());
     });
     let labels = values
         .iter()
-        .map(|(k, v)| {
-            format!(
-                "{}=\"{}\"",
-                k,
-                v.replace('\\', "\\\\")
-                    .replace('"', "\\\"")
-                    .replace('\n', "\\n")
-            )
-        })
+        .map(|(k, v)| format!("{}=\"{}\"", sanitize_label_key(k), sanitize_label_value(v)))
         .collect();
 
     (name, labels)
@@ -319,7 +291,8 @@ fn write_help_line(buffer: &mut String, name: &str, desc: &str) {
     buffer.push_str("# HELP ");
     buffer.push_str(name);
     buffer.push(' ');
-    buffer.push_str(desc);
+    let desc = sanitize_description(desc);
+    buffer.push_str(&desc);
     buffer.push('\n');
 }
 
@@ -345,7 +318,7 @@ fn write_metric_line<T, T2>(
     buffer.push_str(name);
     if let Some(suffix) = suffix {
         buffer.push('_');
-        buffer.push_str(suffix)
+        buffer.push_str(suffix);
     }
 
     if !labels.is_empty() || additional_label.is_some() {
